#!/usr/bin/env python
import os
import sys

import snakemake
import argparse
import yaml

from .version import __title__, __version__
from .reader import AmplimapReaderException, read_new_probe_design, process_probe_design, read_and_convert_mipgen_probes, read_and_convert_heatseq_probes, read_targets, read_snps_txt

def check_config_keys(default_config, my_config, path = []):
    """Recursively check that config keys provided in my_config also exist in default_config (ignoring 'paths' and 'clusters')."""
    differences = []
    for key, value in my_config.items():
        if key in default_config:
            if isinstance(default_config[key], dict) == isinstance(value, dict):
                if isinstance(default_config[key], dict):
                    #both are dicts, keep checking
                    #but don't check paths/cluster because they can contain custom values
                    if not (len(path) == 0 and key in ['paths', 'clusters']):
                        differences += check_config_keys(default_config[key], value, path + [key])
                else:
                    #we're done here
                    pass
            else:
                differences.append(path + [key])
                #raise Exception('Config setting {} is invalid\n'.format(':'.join(path+[key])))
        else:
            differences.append(path + [key])
    return differences

def compare_config_dicts(my_config, used_config, path = []):
    """Recursively search for differences in values between two dicts."""
    differences = []
    for key, value in my_config.items():
        if key in used_config:
            if isinstance(value, dict):
                differences += compare_config_dicts(value, used_config[key], path + [key])
            else:
                if value != used_config[key]:
                    differences.append(path + [key])
        else:
            #key does not exist in used config
            sys.stderr.write('Warning - config key {} not set in used config\n'.format(':'.join(path+[key])))
    return differences

def main(argv = None):
    """
    Run amplimap.
    """
    try:
        basedir = os.path.dirname(os.path.realpath(__file__))
        #sys.stderr.write('Called with arguments: "{}"\n'.format('" "'.join(sys.argv)))
        
        #parse the arguments, which will be available as properties of args (e.g. args.probe)
        parser = argparse.ArgumentParser(
            description = "amplimap v{} - amplicon mapping and analysis pipeline".format(__version__),
            formatter_class = argparse.ArgumentDefaultsHelpFormatter)
        #specify parameters
        parser.add_argument("-v", "--version", help="print version and exit", action="store_true")
        parser.add_argument("--basedir", help="print basedir and exit", action="store_true")
        parser.add_argument("--print-config", help="print configuration (including global and local settings) and exit", action="store_true")
        parser.add_argument("-r", "--run", help="actually run (will perform a dry run otherwise)", action="store_true")
        parser.add_argument("--resume", help="resume analysis in existing analysis directory", action="store_true")
        parser.add_argument("--cluster", help="specify a cluster type defined in your configuration files to run jobs on cluster.")
        parser.add_argument("--skip-file-check", help="skip check for changes in input files when resuming (not recommended)", action="store_true")
        parser.add_argument("--unlock", help="unlock working directory", action="store_true")
        parser.add_argument("--working-directory", help="path to the working directory", default=".")
        parser.add_argument("--ncores", help="number of local cores to run in parallel (only applies if --cluster is NOT set)", default=1, type=int)
        parser.add_argument("--njobs", help="number of cluster jobs to run in parallel (only applies if --cluster is set)", default=10, type=int)
        parser.add_argument("--debug", help="debug mode", action="store_true")
        #parser.add_argument("--debug-dag", help="debug DAG", action="store_true")
        parser.add_argument("TARGET", help="targets to run (eg. pileups variants coverages)", nargs="*")
        if argv is None:
            args = parser.parse_args()
        else:
            args = parser.parse_args(argv)

        if args.debug:
            print('Incoming argv: {}'.format(str(argv)))
            print('Targets: {}'.format(str(args.TARGET)))

        if args.version:
            print('{} {}'.format(__title__, __version__))
            return 0

        if args.basedir:
            print(basedir)
            return 0

        #read default config to get cluster paths etc
        default_config = {}
        default_config_path = os.path.join(basedir, 'config_default.yaml')
        if args.print_config:
            sys.stderr.write('Reading default configuration from: {}\n'.format(default_config_path))
        if os.path.isfile(default_config_path):
            with open(default_config_path, 'r') as config_file:
                default_config = yaml.safe_load(config_file.read())
        else:
            raise Exception('config_default.yaml file missing!')

        #override with data from /etc/amplimap, if exists
        etc_config = {}
        etc_config_path = '/etc/amplimap/%s/config.yaml' % __version__
        if os.path.isfile(etc_config_path):
            if args.print_config:
                sys.stderr.write('Reading additional configuration from: {}\n'.format(etc_config_path))
            with open(etc_config_path, 'r') as config_file:
                etc_config = yaml.safe_load(config_file.read())

        #override with data from $AMPLIMAP_CONFIG, if exists
        env_config = {}
        try:
            env_config_path = os.environ['AMPLIMAP_CONFIG']
            if os.path.isfile(env_config_path):
                if args.print_config:
                    sys.stderr.write('Reading additional configuration from: {}\n'.format(env_config_path))
                with open(env_config_path, 'r') as config_file:
                    env_config = yaml.safe_load(config_file.read())
        except KeyError:
            pass

        #read local config
        local_config = {}
        local_config_path = os.path.join(args.working_directory, 'config.yaml')
        if os.path.isfile(local_config_path):
            with open(local_config_path, 'r') as config_file:
                local_config = yaml.safe_load(config_file.read())
        else:
            sys.stderr.write('No local config.yaml found, using default configuration.\n')

        #merge configs together
        config = default_config
        for my_config in [etc_config, env_config, local_config]:
            #check that all settings actually exist
            differences = check_config_keys(default_config, my_config)
            if len(differences) > 0:
                sys.stderr.write('Your configuration file(s) contain unknown or invalid settings:\n')
                for diff in differences:
                    sys.stderr.write('\t- {}\n'.format(':'.join(diff)))
                sys.stderr.write('Please check their spelling and location and try again.\n')
                return 1

            snakemake.utils.update_config(config, my_config)

        if args.print_config:
            yaml.dump(config, sys.stdout, default_flow_style=False)
            return 0

        #do some basic checks
        assert os.path.isdir(args.working_directory), 'working directory does not exist'
        assert os.path.isdir(os.path.join(args.working_directory, 'reads_in')) \
            or os.path.isdir(os.path.join(args.working_directory, 'bams_in')) \
            or os.path.isdir(os.path.join(args.working_directory, 'tagged_bams_in')) \
            or os.path.isdir(os.path.join(args.working_directory, 'unmapped_bams_in')), 'reads_in/, bams_in/ or unmapped_bams_in/ directory missing'
        # assert os.path.isfile(os.path.join(args.working_directory, 'probes.csv')) \
        #     or os.path.isfile(os.path.join(args.working_directory, 'probes_mipgen.csv')) \
        #     or os.path.isfile(os.path.join(args.working_directory, 'probes_heatseq.tsv')), 'probes.csv, probes_mipgen.csv, or probes_heatseq.tsv file missing'

        #check some basic settings
        aligners = ['naive', 'bwa', 'bowtie2', 'star']
        if not config['align']['aligner'] in aligners:
            raise Exception('align: aligner must be one of {}!'.format(','.join(aligners)))

        callers = ['gatk', 'platypus']
        if not config['variants']['caller'] in callers:
            raise Exception('variants: caller must be one of {}!'.format(','.join(callers)))

        if config['parse_reads']['quality_trim_threshold'] != False:
            if not isinstance(config['parse_reads']['quality_trim_threshold'], float):
                raise Exception('quality_trim_threshold must be a decimal number!')
            if not config['parse_reads']['quality_trim_threshold'] > 0 and config['parse_reads']['quality_trim_threshold'] < 1:
                raise Exception('quality_trim_threshold must be either "false" or above 0 and below 1!')

        if not (config['parse_reads']['min_percentage_good'] >= 0 and config['parse_reads']['min_percentage_good'] <= 100):
            raise Exception('min_percentage_good must be between 0 and 100')

        if not (config['parse_reads']['umi_one'] >= 0 and config['parse_reads']['umi_two'] >= 0):
            raise Exception('umi_one and umi_two must be 0 or greater')

        if config['annotate']['annovar']['protocols'].count(',') != config['annotate']['annovar']['operations'].count(','):
            raise Exception('The number of comma-separated protocols and operations under `annotate: annovar:` must match!')

        #if we don't have UMIs (either on reads or as bam tag) we definitely have to ignore them
        #this makes it possible to "auto-detect" whether we need to ignore_umis or not
        if (config['parse_reads']['umi_one'] + config['parse_reads']['umi_two'] == 0) and config['general']['umi_tag_name'] == "":
            config['general']['ignore_umis'] = True

        #check we have proper paths
        if not config['general']['genome_name'] in config['paths'] or not isinstance(config['paths'][config['general']['genome_name']], dict):
            raise Exception('Could not find list of paths for genome_name: "{}". Please add the paths to your default configuration or your local config.yaml file.'.format(config['general']['genome_name']))

        for name, path in config['paths'][config['general']['genome_name']].items():
            if path.startswith('/PATH/TO/'):
                raise Exception('Path for {} reference is set to {}, which is probably incorrect. Please set the correct path in your default configuration or your local config.yaml file, or leave it empty.'.format(
                    name, path))

        #check input files
        sys.stderr.write('Checking input files...\n')
        if os.path.isfile(os.path.join(args.working_directory, 'probes.csv')):
            read_new_probe_design(os.path.join(args.working_directory, 'probes.csv'), reference_type = 'genome')
        if os.path.isfile(os.path.join(args.working_directory, 'probes_mipgen.csv')):
            process_probe_design(read_and_convert_mipgen_probes(os.path.join(args.working_directory, 'probes_mipgen.csv')))
        if os.path.isfile(os.path.join(args.working_directory, 'probes_heatseq.tsv')):
            process_probe_design(read_and_convert_heatseq_probes(os.path.join(args.working_directory, 'probes_heatseq.tsv')))
        if os.path.isfile(os.path.join(args.working_directory, 'targets.csv')):
            #note: this will fail on overlapping targets
            read_targets(os.path.join(args.working_directory, 'targets.csv'), check_overlaps=True, reference_type = 'genome', file_type = 'csv')
        if os.path.isfile(os.path.join(args.working_directory, 'targets.bed')):
            #note: this will fail on overlapping targets
            read_targets(os.path.join(args.working_directory, 'targets.bed'), check_overlaps=True, reference_type = 'genome', file_type = 'bed')
        if os.path.isfile(os.path.join(args.working_directory, 'snps.txt')):
            read_snps_txt(os.path.join(args.working_directory, 'snps.txt'), reference_type = 'genome')

        #provide our source dir name to snakefile
        config['general']['amplimap_dir'] = basedir
        #this will be used to (very hackily) make sure amplimap can be imported as amplimap.xxx
        #by adding the parent dir to the top of sys.path in the Snakefile
        config['general']['amplimap_parent_dir'] = os.path.dirname(basedir)

        #check if analysis dir exists already
        analysis_dir = os.path.join(args.working_directory, 'analysis')
        configfile = os.path.join(analysis_dir, 'config_used.yaml')
        used_versions_path = os.path.join(analysis_dir, 'versions.yaml')

        #the analysis dir may exist just because we did a dry run, but once the versions exist we actually executed snakemake!
        if os.path.exists(analysis_dir) and os.path.exists(used_versions_path):
            if not args.resume:
                raise Exception('An analysis directory already exists. Please rename it or set --resume to reuse it and possibly overwrite existing files.')
            else:
                #check version
                if os.path.isfile(used_versions_path):
                    with open(used_versions_path, 'r') as used_versions_file:
                        used_versions = yaml.safe_load(used_versions_file.read())
                        if used_versions['_amplimap'] != str(__version__):
                            sys.stderr.write('This analysis was performed with {} {} but this is {} {}!\n\n'.format(__title__, used_versions['_amplimap'], __title__, __version__))
                            sys.stderr.write('Please use the correct version of {} or start a new analysis.\n'.format(__title__))
                            return 1
                        else:
                            sys.stderr.write('{} version checked.\n'.format(__title__))

                #check used config file
                if os.path.isfile(configfile):
                    with open(configfile, 'r') as used_config_file:
                        used_config = yaml.safe_load(used_config_file.read())
                        differences = compare_config_dicts(config, used_config)
                        if len(differences) > 0:
                            sys.stderr.write('config_used.yaml in analysis directory differs from current config.yaml in working directory! Please rename or delete the old analysis directory to restart analysis with the new configuration.\n')
                            sys.stderr.write('Different settings:\n')
                            for diff in differences:
                                sys.stderr.write('\t- {}\n'.format(':'.join(diff)))
                            return 1
                        else:
                            sys.stderr.write('Config files checked.\n')

                #check hashes of input files
                if not args.skip_file_check:
                    used_file_hashes_path = os.path.join(analysis_dir, 'file_hashes.yaml')
                    if os.path.isfile(used_file_hashes_path):
                        with open(used_file_hashes_path, 'r') as used_file_hashes_file:
                            used_file_hashes = yaml.safe_load(used_file_hashes_file.read())

                            from .reader import get_file_hashes
                            for fn, current_hash in get_file_hashes(args.working_directory).items():
                                if used_file_hashes[fn] != current_hash:
                                    sys.stderr.write('File {} seems to have changed since the last run!\n\n'.format(fn))
                                    sys.stderr.write('To ensure consistent results, you should rename or delete the old analysis directory and start a new analysis.\n')
                                    sys.stderr.write('To ignore this error, add the --skip-file-check parameter.\n')
                                    return 1
                            sys.stderr.write('Input files checked.\n')
                else:
                    sys.stderr.write('Warning: Skipping input file check.\n')

        #ensure analysis dir exists now
        try:
            os.makedirs(analysis_dir)
        except OSError as e:
            pass

        #write config to analysis directory, and then use that for snakemake
        with open(configfile, 'w') as f:
            yaml.dump(config, f, default_flow_style=False)

        #set up cluster commands
        cluster_command_nosync = None
        cluster_command_sync = None

        if args.cluster:
            if args.cluster in config['clusters'] and isinstance(config['clusters'][args.cluster], dict):
                if 'command_sync' in config['clusters'][args.cluster]:
                    cluster_command_sync = config['clusters'][args.cluster]['command_sync']
                elif 'command_nosync' in config['clusters'][args.cluster]:
                    cluster_command_nosync = config['clusters'][args.cluster]['command_nosync']
                else:
                    raise Exception('Invalid cluster configuration -- need either command_sync or command_nosync for: {}'.format(args.cluster))
            else:
                raise Exception('Cluster type not found in config: {}'.format(args.cluster))

            sys.stderr.write('Running in cluster mode {} with {} parallel jobs\n'.format(args.cluster, args.njobs))
            sys.stderr.write('cluster_command_nosync={}\n'.format(cluster_command_nosync))
            sys.stderr.write('cluster_command_sync={}\n'.format(cluster_command_sync))

            #make sure cluster log directory exists (this assumed the cluster command is using this as a parameter)
            cluster_logs = os.path.join(args.working_directory, 'cluster_log')
            try:
                os.makedirs(cluster_logs)
            except OSError as e:
                pass
            sys.stderr.write('Will write cluster logs to: {}\n'.format(cluster_logs))
        else:
            sys.stderr.write('Running locally with {} cores\n'.format(args.ncores))

        success = snakemake.snakemake(
            snakefile = os.path.join(basedir, "Snakefile"),
            configfile = configfile,
            cores = args.ncores, #ignored if cluster
            nodes = args.njobs, #ignored if not cluster
            workdir = args.working_directory,
            targets = args.TARGET,
            dryrun = not args.run,
            cluster = cluster_command_nosync,
            cluster_sync = cluster_command_sync,
            jobname = "{}.{{rulename}}.{{jobid}}.sh".format(__title__),
            unlock = args.unlock,
            #debug_dag = args.debug_dag,
            )

        if success:
            if args.unlock:
                sys.stderr.write('Unlocked working directory. Run without --unlock to start.\n')
            elif not args.run:
                sys.stderr.write('{} {} dry run successful. Set --run to run!\n'.format(__title__,  __version__))
            else:
                sys.stderr.write('{} {} finished!\n'.format(__title__, __version__))
            return 0
        else:
            if args.cluster:
                sys.stderr.write('{} {} failed! Please see output above or cluster logs for details.\n'.format(__title__,  __version__))
            else:
                sys.stderr.write('{} {} failed! Please see output above for details.\n'.format(__title__,  __version__))

            return 1
    except AmplimapReaderException as e:
        if args.debug:
            import traceback
            traceback.print_exc()
        sys.stderr.write(str(e))
        sys.stderr.write('{} {} failed!\n'.format(__title__, __version__))
        return 2
    except Exception as e:
        if args.debug:
            import traceback
            traceback.print_exc()
        sys.stderr.write('\nERROR: {}\n\n'.format(e))
        sys.stderr.write('{} {} failed!\n'.format(__title__, __version__))
        return 1

if __name__ == '__main__':
    sys.exit(main())


import re
import sys
import os, hashlib
import pandas as pd
import Bio.Seq

import logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
sh = logging.StreamHandler()
sh.setFormatter(logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))
log.addHandler(sh)

mipgen_columns = ['chr','ext_probe_sequence','lig_probe_sequence','mip_scan_start_position','mip_scan_stop_position','probe_strand','mip_name']
heatseq_columns = ['chromosome','ext_sequence','lig_sequence','target_start','target_stop','probe_strand','probe_id']
probe_columns = ['id', 'first_primer_5to3', 'second_primer_5to3', 'chr', 'target_start', 'target_end', 'strand']

class AmplimapReaderException(Exception):
    """
    Will be raised if reading one of the standard input files has failed, with the aim of providing a more useful error message in the user-visible output.
    """
    def __init__(self, e: Exception, filename: str, should_have_header: bool = None):
        if isinstance(e, ValueError):
            err = 'Invalid value. This usually indicates that there were non-numeric characters (eg. letters, spaces, ...) in a column that should only contain numbers.\n\nMessage: {}'.format(
                str(e)
            )
        elif isinstance(e, AssertionError):
            err = 'Assertion error. This usually indicates that one of the values is invalid or in an unexpected range.\n\nMessage: {}'.format(
                str(e)
            )
        else:
            err = str(e)

        header_note = ''
        if should_have_header is not None:
            if should_have_header:
                header_note = 'Please note that this file SHOULD start with a header line specifying the column names.'
            else:
                header_note = 'Please note that this file SHOULD NOT contain any header line.'

        #add better message to exception
        message = '\n\nError while reading file {}:\n{}\n\n{}\n\n'.format(
            filename,
            err,
            header_note
        )

        #init base exception
        super().__init__(message)

def get_code_versions(path: str = '.') -> dict:
    """
    Get the file modification times of common code files for versioning.
    """
    VERSIONS = {}
    VERSIONS['_amplimap'] = str(__version__)
    VERSIONS['Snakefile'] = str(os.path.getmtime(os.path.join(path, 'Snakefile')))
    VERSIONS['parse_reads'] = str(os.path.getmtime(os.path.join(path, 'parse_reads.py')))
    VERSIONS['pileup'] = str(os.path.getmtime(os.path.join(path, 'pileup.py')))
    VERSIONS['stats_alignment'] = str(os.path.getmtime(os.path.join(path, 'stats_alignment.py')))
    VERSIONS['coverage'] = str(os.path.getmtime(os.path.join(path, 'coverage.py')))
    VERSIONS['simulate'] = str(os.path.getmtime(os.path.join(path, 'simulate.py')))
    VERSIONS['variants'] = str(os.path.getmtime(os.path.join(path, 'variants.py')))
    return VERSIONS

def get_file_hashes(path: str = '.') -> dict:
    """
    Get SHA256 hashes for common input files, so we can make sure these didn't change between runs.
    """
    hashes = {}
    for fn in ['probes.csv', 'targets.bed', 'targets.csv', 'snps.txt']:
        fpath = os.path.join(path, fn)
        if os.path.isfile(fpath):
            file_hash = hashlib.sha256()
            with open(fpath, 'rb', buffering = 0) as file:
                for chunk in iter(lambda : file.read(128*1024), b''):
                    file_hash.update(chunk)
            hashes[fn] = file_hash.hexdigest()
        else:
            hashes[fn] = ''

    return hashes

def merge_probes_by_id(rows: pd.DataFrame) -> pd.Series:
    """
    Merge multiple probes.csv rows with the same probe ID together, to handle cases where MIPGEN generated multiple version because of a SNP.
    """
    #grab first row as template (use .loc instead of .iloc?)
    first_row = rows.iloc[0].copy()

    first_row['n_merged'] = len(rows)
    first_row['n_wildcards'] = None

    if len(rows) > 1:
        first_row['n_wildcards'] = 0

        #make sure these are all the same
        assert (rows['chr'] == first_row['chr']).all(), 'found multiple rows with same probe name but different chr: %s' % first_row['id']
        assert (rows['target_start'] == first_row['target_start']).all(), 'found multiple rows with same probe name but different target_start: %s' % first_row['id']
        assert (rows['target_end'] == first_row['target_end']).all(), 'found multiple rows with same probe name but different target_end: %s' % first_row['id']
        assert (rows['strand'] == first_row['strand']).all(), 'found multiple rows with same probe name but different strand: %s' % first_row['id']

        #merge primers together
        for primer_col in ['first_primer_5to3', 'second_primer_5to3']:
            assert (rows[primer_col].str.len() == len(first_row[primer_col])).all(), 'found multiple rows with same probe id but primer sequences were different lengths'
            for ixrow in range(1, len(rows)):
                for ixchar in range(len(first_row[primer_col])):
                    #if this row doesn't match the first row's sequence
                    if first_row[primer_col][ixchar] != rows.iloc[ixrow][primer_col][ixchar]:
                        #replace character by dot, but only if it hasn't been replaced yet
                        if first_row[primer_col][ixchar] != '.':
                            first_row['n_wildcards'] += 1
                            first_row[primer_col] = first_row[primer_col][:ixchar] + '.' + first_row[primer_col][(ixchar+1):]

        #make sure we actually replaced something with a wildcard
        assert first_row['n_wildcards'] > 0, 'found multiple rows with same probe name but primer sequences were identical: %s' % first_row['id']
        assert first_row['n_wildcards'] < 10, 'found too many differences between primer sequences: %s' % first_row['id']

    del first_row['id'] #remove this, will be attached again later
    return first_row

def read_and_convert_mipgen_probes(path: str) -> pd.DataFrame:
    """
    Read probes file from MIPGEN in CSV format and generate an amplimap probes.csv from it.
    """
    try:
        design = pd.read_csv(path)

        log.info('Read mipgen table from %s -- found %d probes', path, len(design))
        original_len = len(design)
        design.dropna(axis=0, how='all', inplace=True)
        if original_len != len(design):
            log.info('%d probes after dropping empty rows', len(design))
        assert len(design) > 0, 'mipgen table is empty!'

        #make sure we have all the columns we expect
        for column in mipgen_columns:
            if not column in design.columns:
                raise Exception('invalid mipgen table: column %s is missing!' % column)

        #rename columns
        design.rename(columns = {'mip_scan_start_position': 'target_start', 'mip_scan_stop_position': 'target_end', 'probe_strand': 'strand', 'mip_name': 'id'}, inplace=True)

        #in smMIPs, the first arm (read1) seems to be the reverse primer, which would normally be the second arm
        #its sequence in the table seems to be on the forward strand, which needs to be RC'd to get the actual read sequence (5p->3p)
        design['first_primer_5to3'] = [str(Bio.Seq.Seq(s).reverse_complement()) for s in design['lig_probe_sequence'].str.upper()]
        design['second_primer_5to3'] = design['ext_probe_sequence'].str.upper()

        #detect SNPs and replace them by dot
        original_len = len(design)
        design = design.groupby('id').apply(merge_probes_by_id)
        if original_len != len(design):
            log.info('%d/%d probes left after merging probes by ID', len(design), original_len)
        #reset index to turn id into column again
        design.reset_index(inplace=True)  

        #reorder and extract columns
        design = design[probe_columns]
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = True).with_traceback(sys.exc_info()[2])

    return design

def read_and_convert_heatseq_probes(path: str) -> pd.DataFrame:
    """
    UNTESTED: Read probes file from Roche heatseq in CSV format and generate an amplimap probes.csv from it.
    """
    try:
        design = pd.read_csv(path, sep='\t', comment='#')

        log.info('Read heatseq table from %s -- found %d probes', path, len(design))
        original_len = len(design)
        design.dropna(axis=0, how='all', inplace=True)
        if original_len != len(design):
            log.info('%d probes after dropping empty rows', len(design))
        assert len(design) > 0, 'heatseq table is empty!'

        #make sure we have all the columns we expect
        for column in heatseq_columns:
            if not column in design.columns:
                raise Exception('invalid heatseq table: column %s is missing!' % column)

        #rename columns
        design.rename(columns = {'chromosome': 'chr', 'target_stop': 'target_end', 'probe_strand': 'strand', 'probe_id': 'id'}, inplace=True)

        #In this protocol, read1 should contain extension primer and read2 the ligation primer
        #Note that this is the opposite of smMIPs!
        design['second_primer_5to3'] = [str(Bio.Seq.Seq(s).reverse_complement()) for s in design['lig_sequence'].str.upper()]
        design['first_primer_5to3'] = design['ext_sequence'].str.upper()

        #detect SNPs and replace them by dot
        original_len = len(design)
        design = design.groupby('id').apply(merge_probes_by_id)
        if original_len != len(design):
            log.info('%d/%d probes left after merging probes by ID', len(design), original_len)
        #reset index to turn id into column again
        design.reset_index(inplace=True)  

        #reorder and extract columns
        design = design[probe_columns]
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = True).with_traceback(sys.exc_info()[2])

    return design

def read_new_probe_design(path: str, reference_type: str = 'genome') -> pd.DataFrame:
    """
    Read amplimap probes.csv file and return pandas dataframe.
    """
    try:
        design = pd.read_csv(path)

        log.info('Read probe design table from %s -- found %d probes', path, len(design))
        if list(design.columns) == mipgen_columns:
            #NB: smmip data seems to be in F2R1 orientation (second read = fwd in genomic coordinates) for fwd probes
            #but F1R2 orientation (second read = rev) for rev probes.
            #cs-tag data seems to be in F1R2 orientation for fwd targets. unclear for rev targets, but presumably F2R1?
            #in other words, CS-tag is in gene orientation, while smMIP is in opposite
            #so both are swapped for MIPs.
            #is this why sequences in probes.csv are currently so confusing?
            log.info('Detected old MIPGEN format, converting...')

            #read the probes file again in old mipgen format and convert
            design = read_and_convert_mipgen_probes(path)

        design = process_probe_design(design, reference_type)
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = True).with_traceback(sys.exc_info()[2])

    return design

def process_probe_design(design: pd.DataFrame, reference_type: str = 'genome') -> pd.DataFrame:
    """
    Read amplimap probes.csv file and return pandas dataframe.
    """
    #now drop empty rows
    original_len = len(design)
    design.dropna(axis=0, how='all', inplace=True)
    if original_len != len(design):
        log.info('%d probes after dropping empty rows', len(design))
    assert len(design) > 0, 'Probe design table is empty!'

    #make sure we have all the columns we expect
    for column in probe_columns:
        if not column in design.columns:
            raise Exception('invalid probe design table: column %s is missing!' % column)

    #fix data types
    design = design.astype(
        {
            'id': str,
            'chr': str,
            'strand': str,
            'first_primer_5to3': str,
            'second_primer_5to3': str,
            #these need to be int instead of uint, otherwise they'll get turned into floats when subtracting an int (as we do below)
            'target_start': int,
            'target_end': int,
        },
        errors = 'raise'
    )

    #calculate some additional columns
    design.loc[design['strand'] == '+', 'probe_start'] = design.loc[design['strand'] == '+', 'target_start'] - design.loc[design['strand'] == '+', 'first_primer_5to3'].str.len()
    design.loc[design['strand'] == '-', 'probe_start'] = design.loc[design['strand'] == '-', 'target_start'] - design.loc[design['strand'] == '-', 'second_primer_5to3'].str.len()
    design.loc[design['strand'] == '+', 'probe_end'] = design.loc[design['strand'] == '+', 'target_end'] + design.loc[design['strand'] == '+', 'second_primer_5to3'].str.len()
    design.loc[design['strand'] == '-', 'probe_end'] = design.loc[design['strand'] == '-', 'target_end'] + design.loc[design['strand'] == '-', 'first_primer_5to3'].str.len()
    design['target_start_0'] = design['target_start'] - 1 #we have 1-based coords now
    design['probe_start_0'] = design['probe_start'] - 1 #we have 1-based coords now
    design['capture_size'] = design['probe_end'] - design['probe_start_0']
    design['target_length'] = design['target_end'] - design['target_start_0']

    invalid_coords = ~(design['target_length'] > 0)
    if invalid_coords.any():
        print(design.loc[invalid_coords, ['id', 'chr', 'probe_start', 'probe_end', 'strand']])
        raise Exception('Encountered probe regions with invalid target coordinates. Please check the input file and the output above. Note that the probe_end position must always be greater than the probe_start.')

    # #chr should start with chr
    # if reference_type == 'genome':
    #     design.loc[~design['chr'].str.startswith('chr'), 'chr'] = ['chr' + c for c in design.loc[~design['chr'].str.startswith('chr'), 'chr']]

    #check that probe ids don't contain unexpected characters (space breaks bam files for example, since umi will be added as tag)
    assert design['id'].str.contains(r'^[a-zA-Z0-9._:+-]*$').all(), 'probe names may only contain alphanumeric characters and ._:+-'

    #check for dupes
    any_dup = design.duplicated(subset = ['chr', 'probe_start', 'probe_end', 'strand'], keep = False)
    if any_dup.any():
        print('These lines in the design are duplicated:')
        print(design.loc[any_dup, ['chr', 'probe_start', 'probe_end', 'id', 'strand']])

        dup = design.duplicated(subset = ['chr', 'probe_start', 'probe_end', 'strand'], keep = 'first')
        log.warn('Dropping %d/%d duplicated rows', dup.sum(), len(design))
        design = design[~dup]
        log.warn('Now have %d rows after dropping duplicates', len(design))

        raise Exception('Stopping because of duplicate rows in probes.csv!')

    #make sure probes are uppercase
    design['first_primer_5to3'] = design['first_primer_5to3'].str.upper()
    design['second_primer_5to3'] = design['second_primer_5to3'].str.upper()

    #check strand
    if 'strand' in design.columns:
        assert (design['strand'].isin(['+', '-'])).all(), 'strand must be + or -'

    #get reverse complement of sequence for smart trimming and matching to arms reverse read
    for c in ['first_primer_5to3', 'second_primer_5to3']:
        design['%s__rc' % c] = [str(Bio.Seq.Seq(s).reverse_complement()) for s in design[c]]

    #set the index to the probe id (useful for probe stats)
    design.set_index('id', drop = False, inplace = True, verify_integrity = True)

    return design

def read_sample_info(path: str) -> pd.DataFrame:
    """
    Read amplimap sample_info.csv file and return pandas dataframe.
    """
    try:
        sample_infos = pd.read_csv(path, dtype = 'object')

        assert not sample_infos.columns.str.contains('^Unnamed').any(), \
            'Missing column names in sample info table! Please provide names for all columns in the first row.'

        log.info('Read sample info table from %s -- found %d rows', path, len(sample_infos))
        sample_infos.dropna(axis=0, how='all', inplace=True)
        log.info('%d after dropping empty rows', len(sample_infos))

        assert sample_infos.columns[0] == 'Sample' and sample_infos.columns[1] == 'Targets', \
            'Missing Sample/Targets columns! Got %d columns: %s' % (len(sample_infos.columns), '/'.join(sample_infos.columns))

        #split up Targets column into list
        sample_infos['Target'] = sample_infos['Targets'].str.split(';')

        #make new data frame by stacking the Target lists
        #inspired by: https://stackoverflow.com/questions/38372016/split-nested-array-values-from-pandas-dataframe-cell-over-multiple-rows
        non_target_columns = [c for c in sample_infos.columns if c != 'Target']
        sample_infos.set_index(non_target_columns, inplace = True) #this needs need to be all unstacked columns
        #stack each column (only Target actually)
        sample_infos = sample_infos.apply(lambda col: col.apply(pd.Series).stack(), axis=0) 
        #turn multilevel index back into columns
        sample_infos.reset_index(inplace = True)
        #drop Targets column and target index column generated by reset_index
        sample_infos.drop(['Targets', 'level_%d' % len(non_target_columns)], axis=1, inplace = True)

        #check for dupes
        if sample_infos.duplicated(['Sample', 'Target']).any():
            print(targets[sample_infos.duplicated(['Sample', 'Target'], keep=False)])
            raise Exception('Invalid sample_infos table -- found duplicated Sample/Target IDs!')

        #set proper index
        sample_infos.set_index(['Sample', 'Target'], drop = True, inplace = True, verify_integrity = True)

        print(sample_infos.head(3))
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = True).with_traceback(sys.exc_info()[2])

    return sample_infos

def read_targets(path: str, check_overlaps: bool = False, reference_type: str = 'genome', file_type: str = 'bed') -> pd.DataFrame:
    """
    Read amplimap targets.csv or targets.bed file and return pandas dataframe.
    """
    try:
        if file_type == 'bed':
            skiprows = 0
            with open(path, 'r') as f:
                for line in f:
                    if line.startswith('track') or line.startswith('browser'):
                        skiprows += 1
                    else:
                        break

            if skiprows > 0:
                log.info('Skipping %d browser/track rows from target BED files', skiprows)

            targets = pd.read_table(path, header = None, skiprows = skiprows)
            log.info('Read targets from %s: %d rows', path, len(targets))

            if targets.shape[1] == 6:
                targets.columns = ['chr', 'start_0', 'end', 'id', 'score', 'strand']
                #targets = targets[['chr', 'start_0', 'end', 'id']] #remove score/strand
            elif targets.shape[1] == 4:
                targets.columns = ['chr', 'start_0', 'end', 'id']
            elif targets.shape[1] == 3:
                targets.columns = ['chr', 'start_0', 'end']
                targets['id'] = ['target_%d' % i for i in range(len(targets))]
            else:
                print(targets)
                raise Exception('Invalid target BED file: expected to find either 6, 4 or 3 columns separated by tabs.')
        elif file_type == 'csv':
            targets = pd.read_csv(path)

            #check columns
            if not ('chr' in targets.columns and 'start' in targets.columns and 'end' in targets.columns):
                print(targets)
                raise Exception('Invalid target CSV: first line should contain column names, which must include "chr", "start" and "end".')

            #add id if need be
            if not 'id' in targets.columns:
                targets['id'] = ['target_%d' % i for i in range(len(targets))]

            #fix start and calculate start_0
            targets['start'] = targets['start'].astype(int)
            targets['start_0'] = targets['start'] - 1
        else:
            raise Exception('invalid target file type')

        #fix data types
        targets = targets.astype(
            {
                'id': str,
                'chr': str,
                'start_0': int,
                'end': int,
            },
            errors = 'raise'
        )

        #make sure we have no duplicated ids
        if targets.duplicated('id').any():
            print(targets[targets.duplicated('id', keep=False)])
            raise Exception('Invalid target table: found duplicated target ID(s)!')

        #check for overlaps
        if check_overlaps:
            targets_dict = targets.to_dict()
            for ixrow1 in range(len(targets_dict['chr'])-1):
                for ixrow2 in range(ixrow1+1, len(targets_dict['chr'])):
                    if targets_dict['chr'][ixrow1] == targets_dict['chr'][ixrow2] \
                    and targets_dict['start_0'][ixrow1] <= targets_dict['end'][ixrow2] \
                    and targets_dict['end'][ixrow1] >= targets_dict['start_0'][ixrow2]:
                        print(targets.iloc[ixrow1])
                        print(targets.iloc[ixrow2])
                        raise Exception(
                            'Invalid target table: target regions %s and %s overlap! Please merge them or shorten one of the regions' % (
                                targets_dict['id'][ixrow1],
                                targets_dict['id'][ixrow2]
                            )
                        )

        #chr should always start with chr
        targets['chr_original'] = targets['chr']
        # if reference_type == 'genome':
        #     targets.loc[~targets['chr'].str.startswith('chr'), 'chr'] = ['chr' + c for c in targets.loc[~targets['chr'].str.startswith('chr'), 'chr']]

        targets['length'] = targets['end'] - targets['start_0']

        targets_wrong_length = targets['length'] <= 0
        if targets_wrong_length.any():
            print(targets[targets_wrong_length])
            raise Exception('Invalid target table: found targets with length <= 0! Please check start and end coordinates.')

        targets['type'] = 'target'
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = file_type == 'csv').with_traceback(sys.exc_info()[2])

    return targets

def write_targets_bed(path: str, targets: pd.DataFrame):
    """
    Write targets dataframe to bed file.
    """
    if not 'score' in targets.columns:
        targets['score'] = '.'
    if not 'strand' in targets.columns:
        targets['strand'] = '.'

    targets[['chr_original', 'start_0', 'end', 'id', 'score', 'strand']].to_csv(path, header = False, index = False, sep = '\t')

def read_snps_txt(path: str, reference_type: str = 'genome') -> pd.DataFrame:
    """
    Read amplimap snps.txt file and return pandas dataframe.
    """
    try:
        log.info('Loading SNPs from %s', path)
        snps = pd.read_table(path, header = None) #pos is 1-based!

        #detect potential issues with spaces instead of tabs
        if snps.shape[1] == 1:
            if ' ' in snps.iloc[0, 0]:
                log.info('SNP table does not seem to be tab-separated, trying again with spaces')
                snps = pd.read_table(path, header = None, sep='\s+')

        if snps.shape[1] == 5:
            snps.columns = ['chr', 'pos', 'id', 'snp_ref', 'snp_alt']
        elif snps.shape[1] == 4:
            snps.columns = ['chr', 'pos', 'id', 'snp_ref']
        elif snps.shape[1] == 3:
            snps.columns = ['chr', 'pos', 'id']
        elif snps.shape[1] == 2:
            snps.columns = ['chr', 'pos']
            snps['id'] = ['snp_%d' % i for i in range(len(snps))]
        else:
            print(snps)
            raise Exception('Invalid SNP table -- expected to find between 2 and 5 columns.')

        #fix data types
        snps = snps.astype(
            {
                'id': str,
                'chr': str,
                'pos': int,
            },
            errors = 'raise'
        )

        # #chr should always start with chr    
        # if reference_type == 'genome':
        #     snps.loc[~snps['chr'].str.startswith('chr'), 'chr'] = ['chr' + c for c in snps.loc[~snps['chr'].str.startswith('chr'), 'chr']]

        #do we have genotypes for a snp?
        snps['snp_has_genotypes'] = False
        if 'snp_ref' in snps and 'snp_alt' in snps:
            snps.loc[snps['snp_ref'].notnull() & snps['snp_alt'].notnull(), 'snp_has_genotypes'] = True

        #stop if we have duplicate chr/pos
        if snps.duplicated(['chr', 'pos']).any():
            print(snps[snps.duplicated(['chr', 'pos'], keep=False)])
            raise Exception('Invalid SNPs table -- found duplicated chromosome/position!')

        #set the index to the snp id (stops if duplicate IDs)
        snps.set_index('id', drop = False, inplace = True, verify_integrity = True)

        log.info('Found SNP info for %d SNPs -- shape = %s', len(snps), str(snps.shape))
    except Exception as e:
        raise AmplimapReaderException(e, filename = path,  should_have_header = True).with_traceback(sys.exc_info()[2])

    return snps



"""
This does some basic testing of the whole pipeline. Note this needs to be run through pytest and doesn't use unittest.
"""

#make sure we can import from package directory
import sys, os, re, gzip
packagedir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, packagedir) 

import shutil
import subprocess
import pandas as pd

import amplimap.run

#we need to build the cython file here, since snakemake will call an external python that wouldn't
#inherit pyximport
os.system("cythonize -i {}".format(os.path.join(packagedir, "amplimap", "parse_reads_cy.pyx")))

#set config
os.environ['AMPLIMAP_CONFIG'] = os.path.join(packagedir, "sample_data", "config_default.yaml")

def init_wd(path, reads_in_path, umi_one = 0, umi_two = 0):
    assert os.path.isdir(path)

    #remove previous output
    shutil.rmtree(os.path.join(path, 'analysis'), ignore_errors=True)

    #remove previous reads_in and then prepare a new one
    shutil.rmtree(os.path.join(path, 'reads_in'), ignore_errors=True)
    os.mkdir(os.path.join(path, 'reads_in'))

    #turn .fastq files into .fastq.gz
    for file in os.listdir(reads_in_path):        
        if umi_one > 0 and '_R1' in file:
            umi_len = umi_one
        elif umi_two > 0 and '_R2' in file:
            umi_len = umi_two
        else:
            umi_len = 0

        next_umi = None
        with open(os.path.join(reads_in_path, file), 'rt') as fin, gzip.open(os.path.join(path, 'reads_in', '{}.gz'.format(file)), 'wt') as fout:
            for ix, line in enumerate(fin):
                assert line.startswith('@') or ix % 4 != 0
                assert line.startswith('+') or ix % 4 != 2

                if umi_len > 0:
                    if ix % 4 == 0: #name
                        match = re.search(r'_UMI-([^_]+)', line)
                        assert match, 'UMI missing: %s' % line
                        next_umi = match.group(1)
                    elif ix % 4 == 1: #seq
                        line = '{}{}'.format(next_umi[0:umi_len], line)
                    elif ix % 4 == 3: #qual
                        line = '{}{}'.format('A' * umi_len, line)

                fout.write(line)

def test_version(capsys):
    amplimap.run.main(['--version'])
    captured = capsys.readouterr()
    assert captured.out.strip() == '{} {}'.format(amplimap.run.__title__, amplimap.run.__version__)

def test_config(capsys):
    amplimap.run.main(['--print-config'])
    captured = capsys.readouterr()
    assert 'Reading additional configuration from: {}'.format(os.path.join(packagedir, "sample_data", "config_default.yaml")) in captured.err
    # with capsys.disabled():
    #     sys.stdout.write(captured.err)
    #     sys.stdout.write(captured.out)

def check_run(capsys, wd_path):
    #dry-run
    amplimap.run.main(['--working-directory={}'.format(wd_path), 'pileups'])
    captured = capsys.readouterr()
    assert '{} {} dry run successful.'.format(amplimap.run.__title__, amplimap.run.__version__) in captured.err.strip()

    #full run
    amplimap.run.main(['--working-directory={}'.format(wd_path), 'pileups', '--run'])
    captured = capsys.readouterr()
    assert '{} {} finished!'.format(amplimap.run.__title__, amplimap.run.__version__) in captured.err.strip()    

def check_default_stats(wd_path):
    samples = pd.read_csv(os.path.join(wd_path, 'analysis', 'reads_parsed', 'stats_samples.csv'))
    assert len(samples) == 1

    assert samples.loc[0, 'sample'] == 'S1'
    assert samples.loc[0, 'files'] == 2
    assert samples.loc[0, 'pairs_total'] == 7
    assert samples.loc[0, 'pairs_good_arms'] == 6
    assert samples.loc[0, 'pairs_r1_too_short'] == 1

def check_default_pileups(wd_path, expected_coverage = 5, include_too_short = False):
    pileups = pd.read_csv(os.path.join(wd_path, 'analysis', 'pileup', 'pileups_long.csv'))

    #we covered 11bp
    assert len(pileups) == 11

    #we should have 5 reads, except for the raw alignments where we include the pair with short r1/r2
    pileups['expected_coverage'] = expected_coverage
    if include_too_short:
        pileups.loc[(pileups.pos <= 32) | (pileups.pos >= 39), 'expected_coverage'] += 1

    #everything but 35/37 should be ref
    assert pileups.loc[~pileups.pos.isin([35,37]), 'alts'].isnull().all()
    assert (pileups.loc[~pileups.pos.isin([35,37]), 'ref_hq_count'] == pileups.loc[~pileups.pos.isin([35,37]), 'expected_coverage']).all()
    assert (pileups.loc[~pileups.pos.isin([35,37]), 'nonref_hq_count'] == 0).all()

    #only these should be nonref (36 is low-quality in one read)
    assert pileups.loc[pileups.pos.isin([35,37]), 'alts'].notnull().all()
    #one read from L001, one from L002
    assert (pileups.loc[pileups.pos == 35, 'nonref_hq_count'] == 2).all()
    assert (pileups.loc[pileups.pos == 35, 'ref_hq_count'] == pileups.loc[pileups.pos == 35, 'expected_coverage'] - 2).all()
    assert (set(pileups.loc[pileups.pos == 35, 'alts'].iloc[0].split(';')) == set(['A', 'G'])) #explicitly use iloc and no .all() here
    #just one in L001
    assert (pileups.loc[pileups.pos == 37, 'nonref_hq_count'] == 1).all()
    assert (pileups.loc[pileups.pos == 37, 'ref_hq_count'] == pileups.loc[pileups.pos == 37, 'expected_coverage'] - 1).all()
    assert (set(pileups.loc[pileups.pos == 37, 'alts'].iloc[0].split(';')) == set(['G'])) #explicitly use iloc and no .all() here

def test_naive_pileups_simulation(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_naive")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))
    shutil.rmtree(os.path.join(wd_path, 'test__GGCAATATGT_GGCAATCTGT_100'), ignore_errors=True) #make sure this doesn't exist

    #run simulation, replacing A@30 to C
    amplimap.run.main(['--working-directory={}'.format(wd_path), 'test__GGCAATATGT_GGCAATCTGT_100/test_pileups.done', '--run'])
    captured = capsys.readouterr()
    assert '{} {} finished!'.format(amplimap.run.__title__, amplimap.run.__version__) in captured.err.strip()

    pileups = pd.read_csv(os.path.join(wd_path, 'test__GGCAATATGT_GGCAATCTGT_100', 'pileup', 'pileups_long.csv'))    
    assert len(pileups) == 11

    #we should have an A>C SNP at pos 30, in addition to the others
    assert pileups.loc[~pileups.pos.isin([30, 35, 37]), 'alts'].isnull().all()
    assert (pileups.loc[pileups.pos == 30, 'nonref_hq_count'] == 5).all()
    assert (pileups.loc[pileups.pos == 30, 'ref_hq_count'] == 0).all()
    assert (set(pileups.loc[pileups.pos == 30, 'alts'].iloc[0].split(';')) == set(['C'])) #explicitly use iloc and no .all() here


def test_naive_pileups(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_naive")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))

    check_run(capsys, wd_path)
    check_default_stats(wd_path)
    check_default_pileups(wd_path)

def test_bwa_pileups(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_bwa")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))

    check_run(capsys, wd_path)
    check_default_stats(wd_path)
    check_default_pileups(wd_path)

def test_raw_read_pileups(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_bwa_raw")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"))

    check_run(capsys, wd_path)

    #check custom stats
    samples = pd.read_csv(os.path.join(wd_path, 'analysis', 'reads_parsed', 'stats_samples.csv'))
    assert len(samples) == 1

    assert samples.loc[0, 'sample'] == 'S1'
    assert samples.loc[0, 'files'] == 2
    assert samples.loc[0, 'pairs_total'] == 7
    assert samples.loc[0, 'pairs_good_arms'] == 7
    assert samples.loc[0, 'pairs_r1_too_short'] == 0

    check_default_pileups(wd_path, include_too_short = True)

def test_umi_pileups(capsys):
    wd_path = os.path.join(packagedir, "sample_data", "wd_umis")
    init_wd(wd_path, os.path.join(packagedir, "sample_data", "sample_reads_in"), umi_one = 3, umi_two = 4)

    check_run(capsys, wd_path)
    check_default_stats(wd_path)
    check_default_pileups(wd_path, expected_coverage=4) #one less, because two read pairs have same umi!

    #check umi-specific stats
    stats_reads = pd.read_csv(os.path.join(wd_path, 'analysis', 'reads_parsed', 'stats_reads.csv'))
    assert len(stats_reads) == 1
    assert stats_reads.loc[0, 'sample'] == 'S1'
    assert stats_reads.loc[0, 'probe'] == 'Probe1'
    assert stats_reads.loc[0, 'read_pairs'] == 6
    assert stats_reads.loc[0, 'umis_total'] == 5
    assert stats_reads.loc[0, 'umis_coverage_max'] == 2

